/* 
**  mod_smbauth.c -- Apache sample smbauth module
**  [Autogenerated via ``apxs -n smbauth -g'']
**
**  To play with this sample module first compile it into a
**  DSO file and install it into Apache's modules directory 
**  by running:
**
**    $ apxs -c -i mod_smbauth.c
**
**  Then activate it in Apache's httpd.conf file for instance
**  for the URL /smbauth in as follows:
**
**    #   httpd.conf
**    LoadModule smbauth_module modules/mod_smbauth.so
**    <Location /smbauth>
**    SetHandler smbauth
**    </Location>
**
**  Then after restarting Apache via
**
**    $ apachectl restart
**
**  you immediately can request the URL /smbauth and watch for the
**  output of this module. This can be achieved for instance via:
**
**    $ lynx -mime_header http://localhost/smbauth 
**
**  The output should be similar to the following one:
**
**    HTTP/1.1 200 OK
**    Date: Tue, 31 Mar 1998 14:42:22 GMT
**    Server: Apache/1.3.4 (Unix)
**    Connection: close
**    Content-Type: text/html
**  
**    The sample page from mod_smbauth.c
*/ 

/*
 * most of this file was ripped out of other modules, or at least the ideas
 * were taken from them - I think they were:
 *
 *	<httpd-2.0>/modules/aaa/mod_auth.c
 *	<httpd-2.0>/modules/aaa/mod_auth_dbm.c
 *	<mod_auth_external-2.2.1>/mod_auth_external.c
 *	../apache/mod_smbauth.c
 *
 * some (small) bits are my own. Murray.Jensen@csiro.au, 4-Jul-02.
 */

#include "apr_strings.h"

#ifdef GDBM_PASSWD_CACHE
#include <unistd.h>
#include <gdbm.h>
#if defined(SOLARIS2) && SOLARIS2 >= 7
extern char *crypt(const char *, const char *);
#endif
#endif

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"

#include "../samba-shared/valid.h"

module AP_MODULE_DECLARE_DATA smbauth_module;

/*
 * per-directory config info
 */

typedef
    struct {
	int enabled;
	char *domain;
	char *pdc;
	char *bdc;
	char *grpfile;
#ifdef GDBM_PASSWD_CACHE
	char *passwd_cache;
#endif
	int authoritative;
    }
smbauth_dir_config_rec;

static void *
create_smbauth_dir_config(apr_pool_t *p, char *d)
{
    smbauth_dir_config_rec *conf = apr_palloc(p, sizeof(*conf));

    conf->enabled = 0;
    conf->domain = NULL;
    conf->pdc = NULL;
    conf->bdc = NULL;
    conf->grpfile = NULL;
#ifdef GDBM_PASSWD_CACHE
    conf->passwd_cache = NULL;
#endif
    conf->authoritative = 1;  /* fortress is secure by default */

    return conf;
}

static const char *
set_smbauth_slot(cmd_parms *cmd, void *offset, const char *f, const char *t)
{
    if (t && strcmp(t, "smb") != 0) {
        return apr_pstrcat(cmd->pool, "Invalid auth file type: ", t, NULL);
    }

    return ap_set_file_slot(cmd, offset, f);
}

/*
 * per-server (i.e. global) config info
 */

typedef
    struct {
	char *global_domain;
	char *global_pdc;
	char *global_bdc;
	char *global_grpfile;
#ifdef GDBM_PASSWD_CACHE
	char *global_passwd_cache;
	int global_cache_timeout;
#endif
    }
smbauth_server_config_rec;

static void *
create_smbauth_server_config(apr_pool_t *p, server_rec *s)
{
    smbauth_server_config_rec *scr;

    scr = (smbauth_server_config_rec *)apr_palloc(p, sizeof (*scr));

    scr->global_domain = NULL;
    scr->global_pdc = NULL;
    scr->global_bdc = NULL;
    scr->global_grpfile = NULL;
#ifdef GDBM_PASSWD_CACHE
    scr->global_passwd_cache = NULL;
    scr->global_cache_timeout = 8;
#endif

    return (void *)scr;
}

static const char *
set_global_domain(cmd_parms *cmd, void *dummy, const char *value)
{
    smbauth_server_config_rec *scr;
    scr = ap_get_module_config(cmd->server->module_config, &smbauth_module);
    scr->global_domain = (char *)value;
    return NULL;
}

static const char *
set_global_pdc(cmd_parms *cmd, void *dummy, const char *value)
{
    smbauth_server_config_rec *scr;
    scr = ap_get_module_config(cmd->server->module_config, &smbauth_module);
    scr->global_pdc = (char *)value;
    return NULL;
}

static const char *
set_global_bdc(cmd_parms *cmd, void *dummy, const char *value)
{
    smbauth_server_config_rec *scr;
    scr = ap_get_module_config(cmd->server->module_config, &smbauth_module);
    scr->global_bdc = (char *)value;
    return NULL;
}

static const char *
set_global_grpfile(cmd_parms *cmd, void *dummy, const char *value)
{
    smbauth_server_config_rec *scr;
    scr = ap_get_module_config(cmd->server->module_config, &smbauth_module);
    scr->global_grpfile = (char *)value;
    return NULL;
}

#ifdef GDBM_PASSWD_CACHE
static const char *
set_global_passwd_cache(cmd_parms *cmd, void *dummy, const char *value)
{
    smbauth_server_config_rec *scr;
    scr = ap_get_module_config(cmd->server->module_config, &smbauth_module);
    scr->global_passwd_cache = (char *)value;
    return NULL;
}

static const char *
set_global_cache_timeout(cmd_parms *cmd, void *dummy, const char *value) {
  smbauth_server_config_rec *scr;

  scr = ap_get_module_config(cmd->server->module_config, &smbauth_module);
  scr->global_cache_timeout = atol((char *)value);

  return NULL;
}
#endif

/*
 * table of all configuration directives
 */

static const command_rec smbauth_cmds[] =
{
    AP_INIT_TAKE1("SMBAuth_Global_Domain", set_global_domain, NULL,
	RSRC_CONF, "Domain to authenticate under (global default)"),

    AP_INIT_TAKE1("SMBAuth_Global_PDC", set_global_pdc, NULL,
	RSRC_CONF, "Primary Domain Controller (NetBIOS Name) (global default)"),

    AP_INIT_TAKE1("SMBAuth_Global_BDC", set_global_bdc, NULL,
	RSRC_CONF, "Primary Domain Controller (NetBIOS Name) (global default)"),

    AP_INIT_TAKE1("SMBAuth_Global_Groups", set_global_grpfile, NULL,
	RSRC_CONF, "text file containing group names and SMB user IDs "
	"(global default)"),

#ifdef GDBM_PASSWD_CACHE
    AP_INIT_TAKE1("SMBAuth_Global_Passwd_Cache", set_global_passwd_cache, NULL,
	RSRC_CONF, "Path to DBM file containing cached passwords."),

    AP_INIT_TAKE1("SMBAuth_Global_Cache_Timeout", set_global_cache_timeout,
	NULL, RSRC_CONF, "Set global password cache timeout (in hours)."),
#endif

    AP_INIT_FLAG("SMBAuth", ap_set_flag_slot,
	(void *)APR_OFFSETOF(smbauth_dir_config_rec, enabled),
	OR_AUTHCFG, "Enable/Disable SMBAuth"),

    AP_INIT_TAKE1("SMBAuth_Domain", ap_set_string_slot,
	(void *)APR_OFFSETOF(smbauth_dir_config_rec, domain),
	OR_AUTHCFG, "Domain to authenticate under"),

    AP_INIT_TAKE1("SMBAuth_PDC", ap_set_string_slot,
	(void *)APR_OFFSETOF(smbauth_dir_config_rec, pdc),
	OR_AUTHCFG, "Primary Domain Controller (NetBIOS Name)"),

    AP_INIT_TAKE1("SMBAuth_BDC", ap_set_string_slot,
	(void *)APR_OFFSETOF(smbauth_dir_config_rec, bdc),
	OR_AUTHCFG, "Primary Domain Controller (NetBIOS Name)"),

    AP_INIT_TAKE1("SMBAuth_Groups", ap_set_file_slot,
	(void *)APR_OFFSETOF(smbauth_dir_config_rec, grpfile),
	OR_AUTHCFG, "text file containing group names and SMB user IDs"),

    AP_INIT_TAKE12("AuthGroupFile", set_smbauth_slot,
	(void *)APR_OFFSETOF(smbauth_dir_config_rec, grpfile),
	OR_AUTHCFG, NULL),

#ifdef GDBM_PASSWD_CACHE
    AP_INIT_TAKE1("SMBAuth_Passwd_Cache", ap_set_file_slot,
	(void *)APR_OFFSETOF(smbauth_dir_config_rec, passwd_cache),
	OR_AUTHCFG, "Path to DBM file containing cached passwords."),
#endif

    AP_INIT_FLAG("SMBAuth_Authoritative", ap_set_flag_slot,
	(void *)APR_OFFSETOF(smbauth_dir_config_rec, authoritative),
	OR_AUTHCFG, "Set to 'no' to allow access control to be passed along to "
	"lower modules, if the UserID is not known in this module"),

    {NULL}
};

/*
 * group file handling function - ripped almost verbatim from mod_auth.c
 *
 * having a group file means you can authenticate against a central DC
 * to ensure the user is valid, but restrict access to only a subset of
 * those valid accounts (e.g. only those users in your local group - don't
 * want all those weeny PC users having access - do we? :-)
 */

static apr_table_t *
groups_for_user(request_rec *r, char *user, char *grpfile)
{
    apr_pool_t *p = r->pool;
    ap_configfile_t *f;
    apr_table_t *grps = apr_table_make(p, 15);
    apr_pool_t *sp;
    char l[MAX_STRING_LEN];
    const char *group_name, *ll, *w;
    apr_status_t status;

    if ((status = ap_pcfg_openfile(&f, p, grpfile)) != APR_SUCCESS) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                    "Could not open group file: %s", grpfile);
        return NULL;
    }

    apr_pool_create(&sp, p);

    while (!(ap_cfg_getline(l, MAX_STRING_LEN, f))) {
        if ((l[0] == '#') || (!l[0])) {
            continue;
        }
        ll = l;
        apr_pool_clear(sp);

        group_name = ap_getword(sp, &ll, ':');

        while (ll[0]) {
            w = ap_getword_conf(sp, &ll);
            if (!strcmp(w, user)) {
                apr_table_setn(grps, apr_pstrdup(p, group_name), "in");
                break;
            }
        }
    }
    ap_cfg_closefile(f);
    apr_pool_destroy(sp);
    return grps;
}

/* These functions return 0 if client is OK, and proper error status
 * if not... either HTTP_UNAUTHORIZED, if we made a check, and it failed, or
 * HTTP_INTERNAL_SERVER_ERROR, if things are so totally confused that we
 * couldn't figure out how to tell if the client is authorized or not.
 *
 * If they return DECLINED, and all other modules also decline, that's
 * treated by the server core as a configuration error, logged and
 * reported as such.
 */

/* Determine user ID, and check if it really is that user, for HTTP
 * basic authentication...
 */

static int
authenticate_basic_user(request_rec *r)
{
    smbauth_dir_config_rec *conf = (smbauth_dir_config_rec *) \
	ap_get_module_config(r->per_dir_config, &smbauth_module);
    smbauth_server_config_rec *scr = (smbauth_server_config_rec *) \
	ap_get_module_config(r->server->module_config, &smbauth_module);
    const char *sent_pw;
    char *dom, *pdc, *bdc;
    int res;
#ifdef GDBM_PASSWD_CACHE
    char *cache_dbm;
    GDBM_FILE dbf;
    datum package = { r->user, strlen(r->user) + 1 };
#endif

    if ((res = ap_get_basic_auth_pw(r, &sent_pw))) {
        return res;
    }

    if (!conf->enabled) {
        return DECLINED;
    }

    if ((dom = conf->domain) == NULL || *dom == '\0') {
	if ((dom = scr->global_domain) == NULL || *dom == '\0')
	    return DECLINED;
    }

    if ((pdc = conf->pdc) == NULL || *pdc == '\0') {
	if ((pdc = scr->global_pdc) == NULL || *pdc == '\0')
	    return DECLINED;
    }

    if ((bdc = conf->bdc) == NULL) {
	bdc = scr->global_bdc;
    }
    /* bdc is allowed to be unset - but this makes it better for printing */
    if (bdc == NULL)
	bdc = "";

#ifdef GDBM_PASSWD_CACHE
    if ((cache_dbm = conf->passwd_cache) == NULL || *cache_dbm == '\0') {
	cache_dbm = scr->global_passwd_cache;
    }

    if (cache_dbm != NULL && *cache_dbm != '\0') {

	dbf = gdbm_open(cache_dbm, 0, GDBM_READER, 0664, 0);

	if (dbf != NULL) {
	    datum results;

	    results = gdbm_fetch(dbf, package);

	    if (results.dptr != NULL) {
		long now = time(NULL), was;
		char results_str[MAX_STRING_LEN], salt[3];
		char *oldpasswd, *newpasswd;
		int results_len = results.dsize;

		if (results_len >= MAX_STRING_LEN)
		    results_len = MAX_STRING_LEN - 1;
		strncpy(results_str, results.dptr, results_len);
		results_str[results_len] = '\0';
		free(results.dptr);

		if ((oldpasswd = strchr(results_str, ':')) != NULL) {
		    oldpasswd[0] = '\0';

		    salt[0] = oldpasswd[1];
		    salt[1] = oldpasswd[2];
		    salt[2] = '\0';

		    oldpasswd = &oldpasswd[3];

		    was = atol(results_str);

		    if ((was + scr->global_cache_timeout*3600) > now) {
			newpasswd = crypt(sent_pw, salt);
			if (strcmp(oldpasswd, newpasswd) == 0) {
			    gdbm_close(dbf);
			    return OK;
			}
		    }
		    else
			gdbm_delete(dbf, package);
		}
		else
		    gdbm_delete(dbf, package);
	    }

	    gdbm_close(dbf);
	}
    }
#endif

    res = Valid_User(r->user, (char *)sent_pw, pdc, bdc, dom);

    if (res != NTV_NO_ERROR) {
        if (!(conf->authoritative)) {
            return DECLINED;
        }
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
	    "user %s failed validation (pdc=%s,bdc=%s,dom=%s) - error code %d",
	    r->user, pdc, bdc, dom, res);
        ap_note_basic_auth_failure(r);
	switch (res) {
	case NTV_SERVER_ERROR: res = HTTP_SERVICE_UNAVAILABLE; break;
	case NTV_PROTOCOL_ERROR: res = HTTP_INTERNAL_SERVER_ERROR; break;
	case NTV_LOGON_ERROR: res = HTTP_UNAUTHORIZED; break;
	default: res = HTTP_INTERNAL_SERVER_ERROR; break;
	}
	return res;
    }

#ifdef GDBM_PASSWD_CACHE
    if (cache_dbm != NULL && *cache_dbm != '\0') {

	dbf = gdbm_open(cache_dbm, 0, GDBM_WRCREAT, 0664, 0);

	if (dbf == NULL)
	    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "Found DBM error...");
	else {
	    datum dat;
	    char data_str[MAX_STRING_LEN], *fmt_str;

#if !defined(SUNOS4) && (!defined(SOLARIS2) || SOLARIS2 >= 7)
	    fmt_str = "%ld:69%.*s";
#else
	    fmt_str = "%ld:%.*s";
#endif

	    sprintf(data_str, fmt_str,
		time(NULL), MAX_STRING_LEN-1, crypt(sent_pw, "69"));
	    data_str[MAX_STRING_LEN-1] = '\0';

	    dat.dptr = data_str;
	    dat.dsize = strlen(data_str) + 1;

	    gdbm_store(dbf, package, dat, GDBM_REPLACE);

	    gdbm_close(dbf);
	}
    }
#endif

    return OK;
}

/* Checking ID */

static int
check_user_access(request_rec *r)
{
    smbauth_dir_config_rec *conf = (smbauth_dir_config_rec *) \
	ap_get_module_config(r->per_dir_config, &smbauth_module);
    smbauth_server_config_rec *scr = (smbauth_server_config_rec *) \
	ap_get_module_config(r->server->module_config, &smbauth_module);
    char *user = r->user, *grpfile;
    int m = r->method_number;
    int method_restricted = 0;
    register int x;
    const char *t, *w;
    apr_table_t *grpstatus;
    const apr_array_header_t *reqs_arr = ap_requires(r);
    require_line *reqs;

    /* BUG FIX: tadc, 11-Nov-1995.  If there is no "requires" directive, 
     * then any user will do.
     */
    if (!reqs_arr) {
        return OK;
    }
    reqs = (require_line *)reqs_arr->elts;

    if ((grpfile = conf->grpfile) == NULL)
	grpfile = scr->global_grpfile;

    if (grpfile && *grpfile) {
        grpstatus = groups_for_user(r, user, grpfile);
    }
    else {
        grpstatus = NULL;
    }

    for (x = 0; x < reqs_arr->nelts; x++) {

        if (!(reqs[x].method_mask & (AP_METHOD_BIT << m))) {
            continue;
        }

        method_restricted = 1;

        t = reqs[x].requirement;
        w = ap_getword_white(r->pool, &t);
        if (!strcmp(w, "valid-user")) {
            return OK;
        }
        if (!strcmp(w, "user")) {
            while (t[0]) {
                w = ap_getword_conf(r->pool, &t);
                if (!strcmp(user, w)) {
                    return OK;
                }
            }
        }
        else if (!strcmp(w, "group")) {
            if (!grpstatus) {
                return DECLINED;        /* DBM group?  Something else? */
            }

            while (t[0]) {
                w = ap_getword_conf(r->pool, &t);
                if (apr_table_get(grpstatus, w)) {
                    return OK;
                }
            }
        }
        else if (conf->authoritative) {
            /* if we aren't authoritative, any require directive could be
             * valid even if we don't grok it.  However, if we are 
             * authoritative, we can warn the user they did something wrong.
             * That something could be a missing "AuthAuthoritative off", but
             * more likely is a typo in the require directive.
             */
            ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                          "access to %s failed, reason: unknown require "
                          "directive:\"%s\"", r->uri, reqs[x].requirement);
        }
    }

    if (!method_restricted) {
        return OK;
    }

    if (!(conf->authoritative)) {
        return DECLINED;
    }

    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                  "access to %s failed, reason: user %s not allowed access",
                  r->uri, user);
        
    ap_note_basic_auth_failure(r);
    return HTTP_UNAUTHORIZED;
}

static int smbauth_init_handler(apr_pool_t *p, apr_pool_t *plog, 
					apr_pool_t *ptemp, server_rec *s)
{
    ap_add_version_component(p, "mod_smbauth/1.4.3");
    return OK;
}

static void
register_hooks(apr_pool_t *p)
{
    ap_hook_post_config(smbauth_init_handler, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_check_user_id(authenticate_basic_user, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_auth_checker(check_user_access, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA smbauth_module =
{
    STANDARD20_MODULE_STUFF,
    create_smbauth_dir_config,    /* dir config creater */
    NULL,                         /* dir merger --- default is to override */
    create_smbauth_server_config, /* server config */
    NULL,                         /* merge server config */
    smbauth_cmds,                 /* command apr_table_t */
    register_hooks                /* register hooks */
};
